/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2003, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __IMAGIZEAPP__
#include "imagizeapp.h"
#endif

#ifndef __WINUTILS__
#include "winutils.h"
#endif

#ifndef __TRANSFORM__
#include "transform.h"
#endif

#include <cmath>

#include <wingdi.h>

#include "resource.h"

const char *kAppName = "Imagize";

const double PI = 3.14159265358979323846;
const double kZoomFactor = 1.5;
const double kRotateBy = PI / 16;
const double kSkewFactor = 1.25;
const double kPixelsPad = 3.0;

// this is the Window Proc for our main window
long CALLBACK MainWindowProc( HWND hwnd,   UINT msg,
                         UINT wParam, LONG lParam );

// this is the Window Proc for our bitmap window
long CALLBACK BitmapWindowProc( HWND hwnd,   UINT msg,
                         UINT wParam, LONG lParam );


BOOL CALLBACK MainDlgProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);

BOOL CALLBACK AboutDialogProc( HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

ImagizeApp *ImagizeApp::sApp = NULL;

ImagizeApp::ImagizeApp(void)
{
	sApp = this;

	mMainWindow = 0;
	mMainDlg = 0;
	mBitmapWindow = 0;

	mDragging = false;


	mTranslateX = 0.0;
	mTranslateY = 0.0;
	mScaleX = 1.0;
	mScaleY = 1.0;
	mRotateAngle = 0.0;
}

ImagizeApp::~ImagizeApp()
{
	if( mMainWindow )
		DestroyWindow(mMainWindow);

	mMainWindow = NULL;
}

// set transform to identity.
void ImagizeApp::Identity(void)
{
	mTranslateX = 0.0;
	mTranslateY = 0.0;
	mScaleX = 1.0;
	mScaleY = 1.0;
	mRotateAngle = 0.0;
	
	Transform();
	InvalidateRect(mBitmapWindow,NULL,false);
}


bool ImagizeApp::Start(void)
{

	// create a main window class with the given name and window proc
	// and show the window and run the message loop.
	WNDCLASS wc;

	// register our window class
	wc.style = CS_HREDRAW | CS_VREDRAW; // what are these flags?
	wc.lpfnWndProc = MainWindowProc;

	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = mInstance;

	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION );
	wc.hCursor = LoadCursor(NULL, IDC_ARROW );
	wc.hbrBackground = (HBRUSH)GetStockObject(  WHITE_BRUSH );

	wc.lpszMenuName = MAKEINTRESOURCE( IDR_MENU1 );
	wc.lpszClassName = kAppName;

	if( !RegisterClass( &wc ) )
		return false;

	
	// register our bitmap window class.
	wc.style = CS_HREDRAW | CS_VREDRAW; // what are these flags?
	wc.lpfnWndProc = BitmapWindowProc;

	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = mInstance;

	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW );
	wc.hbrBackground = NULL; // (HBRUSH)GetStockObject(  WHITE_BRUSH );

	wc.lpszMenuName = 0; 
	wc.lpszClassName = "ImagizeBitmap";

	if( !RegisterClass( &wc ) )
		return false;





	// and should also be checked against the size of the screen.
	// create our main window
	mMainWindow = CreateWindow( 
		kAppName, // window class name
		kAppName, // caption 
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN ,      // window style            
		CW_USEDEFAULT,            // initial X position      
		CW_USEDEFAULT,            // initial Y position      
		CW_USEDEFAULT,            // initial X size          
		CW_USEDEFAULT,            // initial Y size    
		NULL,                     // parent window handle    
		NULL,                     // set the window menu handle  		
		mInstance,                // program instance handle
		NULL  );                  // creation parameters   

	if( !mMainWindow )
		return false;


	// Load dialog into window
	mMainDlg = CreateDialog(mInstance,MAKEINTRESOURCE(IDD_IMAGIZE),mMainWindow,MainDlgProc);
	if( !mMainDlg )
		return false;

	// get child window.
	mBitmapWindow = GetDlgItem(mMainDlg,IDC_BITMAP);
	if( !mBitmapWindow )
		return false;




	mBitmap.SetWindow( mBitmapWindow );

	mOrigBitmap.CopyResource(mInstance, IDB_BITMAP1);

	// save the offsets of the bitmap window to the dialog window
	// GetWindowRect then ScreenToClient of dialog
	RECT childR, parentR;
	if( !GetClientRect(mMainDlg,&parentR) ||
		!GetChildWindowRect(mMainDlg,mBitmapWindow,&childR) )
		return false;

	mBitmapOffsets.left = childR.left - parentR.left;
	mBitmapOffsets.top = childR.top - parentR.top;
	mBitmapOffsets.right = parentR.right - childR.right;
	mBitmapOffsets.bottom = parentR.bottom - childR.bottom;



	// resize everything to fit the window.
	SizeChanged();


	// load keyboard accelerators.
	mAccel = ::LoadAccelerators(mInstance,MAKEINTRESOURCE(IDR_ACCELERATOR1));
	if( !mAccel )
		return false;

	// show the windows.
	ShowWindow( mMainDlg, true );
	SetFocus( mMainDlg );
	ShowWindow( mMainWindow, true );
	
	return true;
}


void ImagizeApp::Run(void)
{
	if( !Start() )
	{
		// Give MessageBox about failure to startup
		MessageBox(mMainWindow,"Application failed in startup","Startup failed",MB_OK | MB_ICONERROR );
		return;
	}


	// Run message loop ( don't forget IsDialogMessage )

	// Message loop !
	BOOL bRet;
	MSG msg;


	while ( (bRet = GetMessage(&msg, NULL, 0, 0)) != 0 ) 
	{ 
		if (bRet == -1 )
		{
			// handle the error and possibly exit
		}
		else 
		{
			if(!TranslateAccelerator( mMainWindow, mAccel, &msg )) 
			{
				if (!IsWindow(mMainDlg) || !IsDialogMessage(mMainDlg, &msg)) 
				{ 
					TranslateMessage(&msg); 
					DispatchMessage(&msg); 
				}
			}
			else
			{
				DispatchMessage(&msg); 
			}
		}
	}

}



void ImagizeApp::SizeChanged(void) // call when size of main window changed.
{
	RECT r, mr;
	GetClientRect(mMainWindow,&r);
	SetWindowPos(mMainDlg,NULL,r.left,r.top,r.right-r.left,r.bottom-r.top,SWP_NOZORDER);

	GetClientRect(mMainDlg,&r);
	mr = r;
	// inset it the correct amount.
	r.left += mBitmapOffsets.left;
	r.top += mBitmapOffsets.top;
	r.right -= mBitmapOffsets.right;
	r.bottom -= mBitmapOffsets.bottom;
	SetWindowPos(mBitmapWindow,NULL,r.left,r.top,r.right-r.left,r.bottom-r.top,SWP_NOZORDER);

	// set size of bitmap
	long width, height;
	width = r.right - r.left;
	height = r.bottom - r.top;
	// make width & height of bitmap atleast 1 to keep anything bad from happening.
	if( width <= 0 )
		width = 1;
	if( height <= 0 )
		height = 1;

	mBitmap.Resize(r.right-r.left,r.bottom-r.top);

	// Redraw the image.
	Transform();
	InvalidateRect(mBitmapWindow,NULL,false);
	
}

// open bitmap from file
void ImagizeApp::Open(void)
{
	char filename[ MAX_PATH + 5 ]; // just to be sure it can always hold the extension

	// Save a BMP file.
	if( GetImageOpenFileName(mInstance, filename,MAX_PATH + 5) )
	{
		if( !mOrigBitmap.ReadFromFile(filename) )
		{
			MessageBox(mMainWindow,"Error opening image","Error",MB_OK | MB_ICONERROR );
			// reset transformations.

		}
		else
		{
			// set back to identity and redraw.
			Identity();
		}

	}
}


// save bitmap to file
void ImagizeApp::Save(void)
{
	char filename[ MAX_PATH + 5 ]; // just to be sure it can always hold the extension

	// Save a BMP file.
	if( GetImageSaveFileName(mInstance, filename,MAX_PATH + 5) )
	{
		if( !mBitmap.WriteToFile(filename) )
			MessageBox(mMainWindow,"Error saving image","Error",MB_OK | MB_ICONERROR );
	}
}


// transform mOrigBitmap into mBitmap
void ImagizeApp::Transform(void)
{
	Matrix3By3 scaleM, rotateM, translateM, compoundM, tempM;

	long srcWidth, srcHeight, dstWidth, dstHeight;
	srcWidth = mOrigBitmap.GetWidth();
	srcHeight = mOrigBitmap.GetHeight();
	dstWidth = mBitmap.GetWidth();
	dstHeight = mBitmap.GetHeight();

	// set background to black.
	mBitmap.Solid( WBmp::MakeLongColor( 0,0,0 ) );

	TranslationMatrix( translateM, -mTranslateX, -mTranslateY );

	// inverse of scale to get inverse matrix.
	ScalingMatrix( scaleM, (double)srcWidth/2.0, (double)srcHeight/2.0, 1.0 / mScaleX, 1.0 / mScaleY );

	// rotate by negative for inverse
	RotationMatrix( rotateM, (double)srcWidth/2.0, (double)srcHeight/2.0, -mRotateAngle );

	MatrixMultiply( scaleM, rotateM, tempM );
	MatrixMultiply( tempM, translateM, compoundM );

	// Calculate the actual rectangle that will contain the entire image after
	// scaling, rotation, and translation, then use that as bounds
	// for the inverse transformation, instead of the entire thing.
	double sinT = sin( mRotateAngle );
	double cosT = cos( mRotateAngle );

	long left, right, top, bottom, oHeight, oWidth;

	// add a few pixels of padding to be sure we don't clip something due to rounding
	oHeight = (long)((srcHeight*fabs(cosT) + srcWidth*fabs(sinT)) * mScaleY + kPixelsPad);
	oWidth = (long)((srcWidth*fabs(cosT) + srcHeight*fabs(sinT)) * mScaleX + kPixelsPad);

	left = srcWidth/2 - oWidth/2;
	top = srcHeight/2 - oHeight/2;
	right = left + oWidth;
	bottom = right + oHeight;
	
	if( left < 0 )
		left = 0;
	if( top < 0 )
		top = 0;

	if( right > dstWidth )
		right = dstWidth;
	if( bottom > dstHeight )
		bottom = dstHeight;

	if( right > left && bottom > top )
	{
		InvTransformImage(
			compoundM,
			(unsigned char *)mOrigBitmap.GetBuffer(),srcWidth,srcHeight,srcWidth*4,
			(unsigned char *)mBitmap.GetBuffer(),dstWidth,dstHeight,dstWidth*4,
			0,0,dstWidth,dstHeight
			);
	}
}


// this blits the bitmap to the screen.
void ImagizeApp::Draw(void)
{
	mBitmap.Draw(); // that's it
}


// show about box
void ImagizeApp::About(void)
{
	// Show about box
	(void)CreateDialog(mInstance,MAKEINTRESOURCE(IDD_ABOUT),mMainWindow,AboutDialogProc);
}


void ImagizeApp::ZoomIn(void)
{
	mScaleX *= kZoomFactor;
	mScaleY *= kZoomFactor;

	Transform();
	InvalidateRect(mBitmapWindow,NULL,false);
}

void ImagizeApp::ZoomOut(void)
{
	mScaleX /= kZoomFactor;
	mScaleY /= kZoomFactor;

	Transform();
	InvalidateRect(mBitmapWindow,NULL,false);
}

// rotate
void ImagizeApp::Clockwise(void)
{
	mRotateAngle += kRotateBy;

	Transform();
	InvalidateRect(mBitmapWindow,NULL,false);
}

void ImagizeApp::CClockwise(void)
{
	mRotateAngle -= kRotateBy;
	
	Transform();
	InvalidateRect(mBitmapWindow,NULL,false);
}

// skew
void ImagizeApp::Taller(void)
{
	mScaleY *= kSkewFactor;

	Transform();
	InvalidateRect(mBitmapWindow,NULL,false);
}

void ImagizeApp::Wider(void)
{
	mScaleX *= kSkewFactor;

	Transform();
	InvalidateRect(mBitmapWindow,NULL,false);
}




// these handle dragging of the image.
void ImagizeApp::StartDragging(int mx, int my)
{
	mDragLastX = mx;
	mDragLastY = my;

	SetCapture(mBitmapWindow);
	mDragging = true;
}

void ImagizeApp::StopDragging(int mx, int my)
{
	KeepDragging(mx,my);
	mDragging = false;

	if(GetCapture() == mBitmapWindow)
		ReleaseCapture();
}

void ImagizeApp::KeepDragging(int mx, int my)
{
	if( mDragging )
	{
		// actually move the bastard graph
		int dx, dy;
		dx = mx - mDragLastX;
		dy = my - mDragLastY;

		mTranslateX += dx;
		mTranslateY += dy;
		Transform();

		// Draw seems to make it more responsive.
	//	InvalidateRect(mBitmapWindow,NULL,false); // or is it better to call Draw()
		Draw();

		mDragLastX = mx;
		mDragLastY = my;

	}
}

void ImagizeApp::Init(HINSTANCE hInstance)
{
	mInstance = hInstance;
}



// this is the Window Proc for our main window
long FAR PASCAL MainWindowProc( HWND hwnd,   UINT msg,
                         UINT wParam, LONG lParam )
{
	HDC hdc;
	PAINTSTRUCT ps;

	switch( msg )
	{
		case WM_CREATE:
			break;

		case WM_ACTIVATE:
			break;
		
		case WM_PAINT:
			// we don't have anything we have to paint by hand from this window proc.
			hdc = BeginPaint(hwnd,&ps);

			// nothing to draw by hand.

			EndPaint(hwnd,&ps);

			return 0;
		case WM_CLOSE:
			PostQuitMessage(0);
			break;
		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		case WM_RBUTTONDOWN: 

			break;
		case WM_SIZE:
			// window is resizing
			ImagizeApp::GetApp()->SizeChanged();
			break;
		case WM_NOTIFY:
			// notify messages, (notification of something by a child who is a common control)
			
			break;
		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDA_OPEN:
				case IDM_OPEN:
					ImagizeApp::GetApp()->Open();
					break;
				case IDA_SAVE:
				case IDM_SAVE:
					ImagizeApp::GetApp()->Save();
					break;
				case IDM_EXIT:
					PostQuitMessage(0);
					break;
				case IDA_IDENTITY:
				case IDM_IDENTITY:
					ImagizeApp::GetApp()->Identity();
					break;
				case IDA_ZOOMIN:
				case IDM_ZOOMIN:
					ImagizeApp::GetApp()->ZoomIn();
					break;
				case IDA_ZOOMOUT:
				case IDM_ZOOMOUT:
					ImagizeApp::GetApp()->ZoomOut();
					break;
				case IDM_CLOCKWISE:
					ImagizeApp::GetApp()->Clockwise();
					break;
				case IDM_CCLOCKWISE:
					ImagizeApp::GetApp()->CClockwise();
					break;
				case IDM_TALLER:
					ImagizeApp::GetApp()->Taller();
					break;
				case IDM_WIDER:
					ImagizeApp::GetApp()->Wider();
					break;
				case IDM_ABOUT:
					ImagizeApp::GetApp()->About();
					break;
			}
			break;

		default:
			return DefWindowProc(hwnd,msg,wParam,lParam) ;
	}

	return 0;
}

BOOL CALLBACK MainDlgProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message) 
    { 
        case WM_INITDIALOG: 
            return TRUE; 
 
        case WM_COMMAND: 
			// handle button presses.
            switch (LOWORD(wParam)) 
            { 	
			case IDC_IDENTITY:
				ImagizeApp::GetApp()->Identity();
				break;
			case IDC_ZOOMIN:
				ImagizeApp::GetApp()->ZoomIn();
				break;
			case IDC_ZOOMOUT:
				ImagizeApp::GetApp()->ZoomOut();
				break;
			case IDC_CLOCKWISE:
				ImagizeApp::GetApp()->Clockwise();
				break;
			case IDC_CCLOCKWISE:
				ImagizeApp::GetApp()->CClockwise();
				break;
			case IDC_TALLER:
				ImagizeApp::GetApp()->Taller();
				break;
			case IDC_WIDER:
				ImagizeApp::GetApp()->Wider();
				break;
            } 
		case WM_KEYDOWN:
			break;
    } 
    return FALSE; 
}



// this is the Window Proc for our main window
long FAR PASCAL BitmapWindowProc( HWND hwnd,   UINT msg,
                         UINT wParam, LONG lParam )
{
	HDC hdc;
	PAINTSTRUCT ps;
	POINT pt;

	switch( msg )
	{
		case WM_CREATE:
			break;
		case WM_PAINT:
			hdc = BeginPaint(hwnd,&ps);
			
			ImagizeApp::GetApp()->Draw();

			EndPaint( hwnd, &ps );
			return 0;
		case WM_CLOSE:
			break;
		case WM_DESTROY:
			break;
		case WM_LBUTTONDOWN:
			pt.x = MAKEPOINTS(lParam).x;
			pt.y = MAKEPOINTS(lParam).y;
			ImagizeApp::GetApp()->StartDragging(pt.x,pt.y);
			break;
		case WM_LBUTTONUP:
			pt.x = MAKEPOINTS(lParam).x;
			pt.y = MAKEPOINTS(lParam).y;
			ImagizeApp::GetApp()->StopDragging(pt.x,pt.y);
			break;
		case WM_MOUSEMOVE:
			pt.x = MAKEPOINTS(lParam).x;
			pt.y = MAKEPOINTS(lParam).y;
			ImagizeApp::GetApp()->KeepDragging(pt.x,pt.y);
			break;
		case WM_RBUTTONDOWN: 
			break;
		default:
			return DefWindowProc(hwnd,msg,wParam,lParam) ;
	}

	return 0;
}



BOOL CALLBACK AboutDialogProc( HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg)
	{
		case WM_INITDIALOG:
		{
			return true; // sets the appropriate focus
		}
		case WM_COMMAND:
		{
			switch(LOWORD(wParam))
			{
				case IDC_WEBPAGE:
					// show our web page
					(void)ShellExecute(NULL, "open", "http://www.ranton.org", NULL, NULL, SW_SHOW);
					break;
				case IDOK:
				case IDCANCEL:
					EndDialog(hwnd, IDOK);
					break;
				default:
					break;
			}
			break;
		}
		default:
			return false;// return zero if we did not process message
	}

	return true; 
}
 